# BeetlSQL，DAO平台设计指南

BeetlSQL设计为一个Dao平台，可以在其搭建SQL和NOSQL 访问工具，如果遵循BeetlSQL的API和机制，其DAO会简单易用，功能强大，并解决现有Dao工具中跨数据库，维护重构难，开发效率低的问题

## 自定义Mapper

使用或者定制BeetlSQL，可以从Mapper开始，BeetlSQL提供了BaseMapper接口，作为推荐的使用BeetlSQL首选方式

```java
public interface UserMapper extends BaseMapper<User>{
  
}
```

这里，我们假定一个User对象，我们希望对其增删改查，对应数据库user表。

> 如果你还不熟悉BeetSQL用法，阅读《BeetlSQL使用指南》



BaseMapper提供了内置的CRUD等若干方法，满足50%的常用DAO操作,选取如下4个方法

```java

public interface BaseMapper<T> {
  
  @AutoMapper(InsertAmi.class)
  void insert(T entity);
  
  @AutoMapper(UpdateByIdAmi.class)
  int updateById(T entity);
  
  @AutoMapper(UniqueAmi.class)
  T unique(Object autoKey);
  
  @AutoMapper(SingleAmi.class)
  T single(Object autoKey);
  
}

```

* insert 插入User到相应的表里
* updateById，更新User，按照主键id作为条件更新
* unique，按照主键查找User，如果未找到，抛出异常
* single，同unique，如果未找到，返回null



这四个内置方法，都使用了注解@AutoMapper，定义如下

```java
@Target({java.lang.annotation.ElementType.METHOD})
@Retention(RetentionPolicy.RUNTIME)
public @interface AutoMapper {
	Class<? extends MapperInvoke> value() ;
}
```



AutoMapper需要提供一个类，必须是`MapperInvoke`的子类，此子类真正负责执行Dao操作。我们以InsertAmi为例子，代码如下

```java
public class InsertAmi implements MapperInvoke {
    @Override
    public Object call(SQLManager sm, Class entityClass, Method m, Object[] args) {
        int ret = sm.insert(args[0]);
        return ret;
    }
}
```



可以看到InsertAmi非常简单，仅仅调用SQLManager的insert方法，SQLManager是BeetlSQL提供的核心类

如果你想定义自己的通用Mapper，可以定义任意一个接口，或者继承已经做好的BaseMapper，使用@AutoMapper标明其实现类即可。



除了AutoMapper注解，BeetlSQL也提供了其他注解，甚至是可以自定义注解，解释如下

BeetlSQL提供了@Sql注解，允许同时提供sql语句

```java

@Sql("select * from beetlSQLSysUser where id= ?)
public User queryById(Integer id);
```

解释@Sql注解的也是一个MapperInvoke子类，位于org.beetl.sql.mapper.ready包下，

```java
public class SelectSQLReadyMI extends BaseSQLReadyMI {
    boolean isSingle = false;
    public SelectSQLReadyMI(String sql,Class targetType,boolean isSingle){
        this.sql = sql;
        this.targetType = targetType;
        this.isSingle = isSingle;
    }

    @Override
    public Object call(SQLManager sm, Class entityClass, Method m, Object[] args) {
        SQLReady sqlReady = new SQLReady(this.getSql(),args);
        List list =  sm.execute(sqlReady,this.getTargetType());
        if(isSingle){
            return list.isEmpty()?null:list.get(0);
        }else{
            return list;
        }
    }
}
```



可以看到，SelectSQLReadyMI的call方法仍然非常简单，调用SQLManager的jdbc查询方法。SelectSQLReadyMI的构造是通过MapperMethodParser.parse方法，稍后会介绍这个类（TODO）。

如果你想利用BeetlSQL实现自定义注解，比如你觉得提供sql语句繁琐，你更喜欢SpringData那种根据方法名来推测sql语句，类似如下

```java
@SpringData
public User queryByNameAndAge(String name,Integer a);
```

> Spring Data 是一个Spring封装的DAO框架，它会根据dao的方法命名来判断,如上方法，表示查询用户(query)，条件是通过name属性，因此sql是 select * from beetlSQLSysUser where name=?  and age=?



为了使BeetlSQL能识别SpringData注解，且能使用MapperInvoke执行，你可以定义SpringData注解，如下

```java
@Target({java.lang.annotation.ElementType.METHOD})
@Retention(RetentionPolicy.RUNTIME)
@Builder(SpringDataBuilder.class)
public @interface SpringData {

}
```



这里，@SpringData注解又使用了BeetlSQL的@Builder注解，这表明@SpringData注解的解释是通过此类来定义的，BeetlSQL期望SpringDataBuilder实现MapperExtBuilder接口,能解析当前方法，返回一个MapperInvoke接口，类似如下

```java
public class SpringDataBuilder  implements MapperExtBuilder {
    @Override
    public MapperInvoke parse(Class entity, Method m) {
        ......
    }
}

```

@Builder注解是BeetlSQL一个很强大的注解，它是注解的注解，表明了注解的含义，我们在后面还能看到很多注解都采用了这个方法，比如Mapper中的SqlProvider, Bean的注解扩展@UpdateTime等等

## DBStyle

DBStyle接口定义了SQL数据库或者兼容JDBC的大数据的方言，用户根据自己的SQL或者NOSQL，可以使用BeetlSQL自带的，也可以做扩展。DBStyle定义如下

```java
/**
 * 用来描述数据库差异，主键生成，sql语句，翻页等
 *
 * @author xiandafu
 */
public interface DBStyle extends DBAutoGeneratedSql {


  

    /**
     * 返回DBStyle名称
     * @return
     */
    String getName();

    /**
     * 返回一个DBStyle代码
     * @return
     */
    int getDBType();

    /**
     * 把正常sql转换成分页sql, 参数是 ?
     * <pre>
     * 正常sql: select * from tb_bee
     *
     * 假设 offset = 2
     * 假设 pageSize  = 9
     * (mysql示例) 出来的sql:
     * select * from tb_bee limit 2 , 9
     * </pre>
     *
     * @param sql      正常查询sql
     * @param offset   offset
     * @param pageSize pageSize
     * @return 分页sql
     */
    String getPageSQLStatement(String sql, long offset, long pageSize);

    /**
     * 得到一个jdbc sql的翻页语句写法
     * @param sql
     * @return
     */
    String getPageSQL(String sql);

    void initPagePara(Map<String, Object> paras, long start, long size);

    int getIdType(Class c,String idProperty);

    KeyWordHandler getKeyWordHandler();
    
    /**
     * 通过序列名字返回获取序列值的sql片段
     * @param seqName
     * @return
     */
    public String getSeqValue(String seqName);

    /**
     * sql中的关键字处理
     * @param keyWordHandler
     */
    void setKeyWordHandler(KeyWordHandler keyWordHandler);



    /**
     * jdbc 的batch 个数在数据库收到限制，beetlsql会把批处理数据分批处理
     * 这里配置了一个数据库允许的最大批处理个数，默认1000个
     * @return
     */
    default int getMaxBatchCount(){
        return 1000;
    }

    /**
     * 是否是NOSql
     * @return
     */
    default boolean  isNoSql(){
        return false;
    }

    /**
     * 对应的数据库是否支持jdbc metadata
     * @return
     */
    default boolean metadataSupport(){
        return true ;
    }

    /**
     * 是否支持PreparedStatement，对于数据库来说，几乎都支持，对于Nosql，则不一定支持
     * @return
     */
    default boolean preparedStatementSupport(){
        return true;
    }

    /**
     * 如果不支持preparedStatement，在直接使用Statement的时候，返回输出变量到sql语句里，例如，当变量是字符串a"bc
     * 应该输出{@code "a\"bc"}
     * @param value
     * @return
     */
    public String wrapStatementValue(Object value);

    /**
     * 得到一个SQL执行类
     * @param executeContext
     * @return
     */
    SQLExecutor buildExecutor(ExecuteContext executeContext);


    MetadataManager initMetadataManager(ConnectionSource cs);

    MetadataManager initMetadataManager(ConnectionSource cs,String defaultSchema,String defalutCatalog);
  
    void init(SQLTemplateEngine sqlTemplateEngine);


}

```



## BeetlSQL NOSQL 列表

* hive
* impala+kudu,
* 